"""The 'worktoy.ezdata' class provides the EZData class!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._field_tags import BeginFields, EndFields
from ._data_namespace import DataNamespace
from ._data_metaclass import DataMetaclass, EZData
