"""WorkToy provides a collection of tools."""
#  AGPL-3.0 license
#  Copyright (c) 2023-2024 Asger Jon Vistisen
from __future__ import annotations

import worktoy.parse
import worktoy.meta
import worktoy.text
import worktoy.desc
import worktoy.keenum
import worktoy.ezdata
import worktoy.yolo
